{% load i18n %}

<h3>{% trans "Table Overview" %}</h3>

<div class="info detail">
  <dl>
    <dt>{{ datasource_type }} {% trans "Data Source" %}</dt>
    <dd>{{ datasource_name }}</dd>
    <dt>{% trans "Name" %}</dt>
    <dd>{{ table_name }}</dd>
    <dt>{% trans "ID" %}</dt>
    <dd>{{ id|default:table_name }}</dd>
  </dl>
</div>
